#include "Effect.h"
#include "DEFINES.h"
#include "Crawler.h"
#include "utils.h"

INCLUDE_game
INCLUDE_MONSTER_LIST

Meteor::Meteor(vf2d pos, float lifetime, AnimationState animation, bool upperLevel, vf2d size, float fadeout, vf2d spd, Pixel col, float rotation, float rotationSpd, bool additiveBlending)
	:Effect(pos,lifetime,animation,upperLevel,size,fadeout,spd,col,rotation,rotationSpd,additiveBlending),startLifetime(lifetime){

}

bool Meteor::Update(float fElapsedTime){
	if(lifetime<=0&&!shakeField){
		shakeField=true;
		game->SetupWorldShake(2);
		vf2d meteorOffset=pos+vf2d{lifetime,-lifetime}*320-vf2d{0,GetFrame().GetSourceRect().size.y/4.f};
		for(int i=0;i<650;i++){
			float randomAngle=util::random(2*PI);
			float randomRange=100*size.x*(1-util::random(0.25))*(1-util::random(0.25));
			float randomColorTintG=256-(util::random(128)+util::random(128));
			float randomColorTint=util::random(128);
			vf2d effectPos=vf2d{cos(randomAngle),sin(randomAngle)}*randomRange+meteorOffset;
			game->AddEffect(std::make_unique<Effect>(effectPos,0,AnimationState::DOT_PARTICLE,OnUpperLevel(),vf2d{util::random(2)+1,util::random(3)+1},util::random(3)+1,vf2d{util::random(10)-5,-util::random(20)-5},Pixel{255,uint8_t(randomColorTintG),uint8_t(randomColorTint),uint8_t(util::random(128)+128)},0,0,true),effectPos.y<meteorOffset.y);
		}
		game->HurtEnemies(pos,4*24,game->GetPlayer()->GetAttack()*9,OnUpperLevel());
		game->AddEffect(std::make_unique<PulsatingFire>(pos,3,AnimationState::FIRE_RING1,OnUpperLevel(),vf2d{8,8},1),true);
	}
	return Effect::Update(fElapsedTime);
}

void Meteor::Draw(){
	if(lifetime>0){
		vf2d scale=vf2d{192,64}/3.f*(startLifetime+1-lifetime)*0.25*size;
		vf2d meteorOffset=vf2d{lifetime,0}*120;
		vf2d centerPoint=pos-vf2d{game->GFX_Circle.Sprite()->width*scale.x/2,game->GFX_Circle.Sprite()->height*scale.y/2};
		game->view.DrawDecal(centerPoint+meteorOffset,game->GFX_Circle.Decal(),scale,{0,0,0,192});
	}
	vf2d meteorOffset=pos+vf2d{lifetime,-lifetime}*320-vf2d{0,GetFrame().GetSourceRect().size.y/4.f}*size;
	if(lifetime<=0){
		meteorOffset=pos-vf2d{0,GetFrame().GetSourceRect().size.y/4.f}*size;
	}
	game->view.DrawPartialRotatedDecal(meteorOffset,GetFrame().GetSourceImage()->Decal(),rotation,GetFrame().GetSourceRect().size/2,GetFrame().GetSourceRect().pos,GetFrame().GetSourceRect().size,size,{col.r,col.g,col.b,uint8_t(fadeout/original_fadeoutTime*255)});
}